//
// Lighthouse3D.com
//
//
// This demo was built for learning purposes only.
// Some code could be severely optimised, but I tried to
// keep as simple and clear as possible.
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//
// If you do use it I would love to hear about it. Just post a comment
// at Lighthouse3D.com

// Have Fun :-)

//
// Program for lecturing purposes by Marek Vinkler (xvinkl@fi.muni.cz).
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//


#include <cmath>
#include <cstdlib>
#include <algorithm>

// Include GLEW to access OpenGL 3.3 functions
#include <GL/glew.h>

// GLUT is the toolkit to interface with the OS
#include <GL/freeglut.h>

// Use Very Simple Libs
#include "vsMathLib.h"
#include "vsShaderLib.h"
#include "vsResSurfRevLib.h"
#include "bloom.h"

//----------------------------------------------------------------
// TODO:
// 1) Inspect functions init, processKeys and renderScene.
//----------------------------------------------------------------

VSMathLib *vsml;
VSShaderLib sceneShader, filterShader, renderShader;
VSResSurfRevLib surfRev;

// Camera Position
float camX, camY, camZ;

// Mouse Tracking Variables
int startX, startY, tracking = 0;

// Camera Spherical Coordinates
float alpha = 0.0f, beta = 75.0f;
float r = 20.0f;

// Geometry offset
float offsetX = 0.f, offsetY = 0.f;

// Light direction
float lightPos[4] = {0.0f, 6.0f, 0.0f, 1.0f};

// Window size
int width;
int height;

// Scene properties
const int numObjects = 32;
const int maxWindowWidth = 2048;
const int maxWindowHeight = 2048;

// Vao
GLuint vao;

// Texture
GLuint sceneTexture;
GLuint highlightTexture;
GLuint depthTexture;
GLuint filterTexture[2];

// Framebuffer
GLuint renderFramebuffer;
GLuint filterFramebuffer[2];

// HDR
float exposure = 2.f;
float sceneFactor = 1.f;
float bloomFactor = 1.f;


// ------------------------------------------------------------
//
// Reshape Callback Function
//

void changeSize(int w, int h)
{
	float ratio;

	// Prevent a divide by zero, when window is too short
	if(h == 0)
		h = 1;

	// Set the viewport to be the entire window
	width = w;
	height = h;
	glViewport(0, 0, w, h);

	ratio = (1.0f * w) / h;
	vsml->loadIdentity(VSMathLib::PROJECTION);
	vsml->perspective(53.13f, ratio, 0.1f, 10000.0f);
}


// ------------------------------------------------------------
//
// Render stuff
//

float rangeRandom(float low, float high)
{
	float rnd = (float)rand();
	rnd = rnd/(float)RAND_MAX; // [0,1] range
	return low + rnd * (high-low); // [low,high] range
}

void renderScene(void)
{
	// Reinit random number generator
	srand(0);

	vsml->loadIdentity(VSMathLib::VIEW);
	vsml->loadIdentity(VSMathLib::MODEL);

	// Set camera
	vsml->lookAt(camX, camY, camZ, 0,0,0, 0,1,0);

	vsml->translate(offsetX, offsetY, 0.f);

	/************************************** Render Scene **************************************/
	glViewport(0, 0, width, height);

	glBindFramebuffer(GL_FRAMEBUFFER, renderFramebuffer);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);

	// Use scene shader
	glUseProgram(sceneShader.getProgramIndex());

	// Transform light to camera space and send it to GLSL
	float res[4];
	vsml->multMatrixPoint(VSMathLib::VIEW, lightPos, res);
	sceneShader.setBlockUniform("Lights", "l_pos", res);

	for(int i = 0 ; i < numObjects; ++i)
	{
		// Equivalent to this, but faster!
		//vsml->pushMatrix(VSMathLib::MODEL);
		//vsml->translate(...);
		//surfRev.render();
		//vsml->popMatrix(VSMathLib::MODEL);

		vsml->loadIdentity((VSMathLib::MatrixTypes)(VSMathLib::AUX0+i));
#if 0
		vsml->translate((VSMathLib::MatrixTypes)(VSMathLib::AUX0+i), rangeRandom(-4.f, 4.f), rangeRandom(-3.f, 3.f), rangeRandom(-3.f, 3.f));
#else
		vsml->translate((VSMathLib::MatrixTypes)(VSMathLib::AUX0+i), rangeRandom(-6.f, 6.f), rangeRandom(-5.f, 5.f), rangeRandom(-5.f, 5.f));
#endif
		vsml->rotate((VSMathLib::MatrixTypes)(VSMathLib::AUX0+i), rangeRandom(0.f, 360.f), rangeRandom(-1.f, 1.f), rangeRandom(-1.f, 1.f), rangeRandom(-1.f, 1.f));
		vsml->setUniformArrayIndexName((VSMathLib::MatrixTypes)(VSMathLib::AUX0+i), "m_model[0]", i);
	}

	surfRev.render(numObjects);

	/************************************** Filter highlights **************************************/

	glDisable(GL_DEPTH_TEST);

	// Use scene shader
	glUseProgram(filterShader.getProgramIndex());
	filterShader.setUniform("highlight", 0);

	// Render VAO
	glBindVertexArray(vao);

	// Blur vertically
	glBindFramebuffer(GL_FRAMEBUFFER, filterFramebuffer[0]);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, highlightTexture);
	glViewport(0, 0, height, width);
	glDrawElements(GL_TRIANGLES, faceCount*3, GL_UNSIGNED_INT, 0);

	// Blur horizontally
	glBindFramebuffer(GL_FRAMEBUFFER, filterFramebuffer[1]);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, filterTexture[0]);
	glViewport(0, 0, width, height);
	glDrawElements(GL_TRIANGLES, faceCount*3, GL_UNSIGNED_INT, 0);

	/************************************** Compose Scene **************************************/

	glUseProgram(renderShader.getProgramIndex());
	renderShader.setUniform("scene", 0);
	renderShader.setUniform("highlight", 1);

	renderShader.setUniform("exposure", exposure);
	renderShader.setUniform("sceneFactor", sceneFactor);
	renderShader.setUniform("bloomFactor", bloomFactor);

	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, sceneTexture);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, filterTexture[1]);
	glDrawElements(GL_TRIANGLES, faceCount*3, GL_UNSIGNED_INT, 0);

	// Swap buffers
	glutSwapBuffers();
}


// ------------------------------------------------------------
//
// Events from the Keyboard
//

void processKeys(unsigned char key, int xx, int yy)
{
	switch(key)
	{
		case 27:
			glutLeaveMainLoop();
			break;

		case '+':
			exposure *= 1.1f;
			break;

		case '-':
			exposure /= 1.1f;

		case 'q':
			sceneFactor += 0.1f;
			break;

		case 'a':
			sceneFactor -= 0.1f;
			break;

		case 'w':
			bloomFactor += 0.1f;
			break;

		case 's':
			bloomFactor -= 0.1f;
			break;
	}

	// Prevent meaningless values
	sceneFactor = std::max(std::min(sceneFactor, 1.f), 0.f);
	bloomFactor = std::max(std::min(bloomFactor, 1.f), 0.f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// ------------------------------------------------------------
//
// Mouse Events
//

void processMouseButtons(int button, int state, int xx, int yy)
{
	// Start tracking the mouse
	if (state == GLUT_DOWN)
	{
		startX = xx;
		startY = yy;
		if (button == GLUT_LEFT_BUTTON)
			tracking = 1;
		else if (button == GLUT_RIGHT_BUTTON)
			tracking = 2;
		else if (button == GLUT_MIDDLE_BUTTON)
			tracking = 3;
	}

	// Stop tracking the mouse
	else if (state == GLUT_UP)
	{
		if (tracking == 1)
		{
			alpha -= (xx - startX);
			beta += (yy - startY);
			if (beta > 85.0f)
				beta = 85.0f;
			else if (beta < -85.0f)
				beta = -85.0f;
		}
		else if (tracking == 2)
		{
			r += (yy - startY) * 0.01f;
			if (r < 0.1f)
				r = 0.1f;
		}
		tracking = 0;
	}
}

// Track mouse motion while buttons are pressed
void processMouseMotion(int xx, int yy)
{
	int deltaX, deltaY;
	float alphaAux, betaAux;
	float rAux;

	deltaX =  - xx + startX;
	deltaY =    yy - startY;

	// Left mouse button: move camera
	if (tracking == 1)
	{
		alphaAux = alpha + deltaX;
		betaAux = beta + deltaY;

		if (betaAux > 85.0f)
			betaAux = 85.0f;
		else if (betaAux < -85.0f)
			betaAux = -85.0f;
		rAux = r;
	}
	// Right mouse button: zoom
	else if (tracking == 2)
	{
		alphaAux = alpha;
		betaAux = beta;
		rAux = r + (deltaY * 0.01f);
		if (rAux < 0.1f)
			rAux = 0.1f;
	}
	// Middle mouse button: move object
	else if (tracking == 3)
	{
		alphaAux = alpha;
		betaAux = beta;
		rAux = r;

		offsetX -= deltaX * 0.0001f;
		offsetY -= deltaY * 0.0001f;
	}

	camX = rAux * sin(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camZ = rAux * cos(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camY = rAux *   						       sin(betaAux * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}

void mouseWheel(int wheel, int direction, int x, int y)
{
	r += direction * 0.1f;
	if (r < 0.1f)
		r = 0.1f;

	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// --------------------------------------------------------
//
// Shader Stuff
//

bool setupShaders()
{
	// Shader for scene
	sceneShader.init();
	sceneShader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/scene.vert");
	sceneShader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/scene.frag");

	// Set semantics for the shader variables
	sceneShader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");
	sceneShader.setVertexAttribName(VSShaderLib::NORMAL_ATTRIB, "normal");

	sceneShader.prepareProgram();
	printf("InfoLog sceneShader\n%s\n\n", sceneShader.getAllInfoLogs().c_str());

	// Shader for filter
	filterShader.init();
	filterShader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/pass.vert");
	filterShader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/filter.frag");

	// Set semantics for the shader variables
	filterShader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");

	filterShader.prepareProgram();
	printf("InfoLog filterShader\n%s\n\n", filterShader.getAllInfoLogs().c_str());

	// Shader for render
	renderShader.init();
	renderShader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/pass.vert");
	renderShader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/render.frag");

	// Set semantics for the shader variables
	renderShader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");

	renderShader.prepareProgram();
	printf("InfoLog renderShader\n%s\n\n", renderShader.getAllInfoLogs().c_str());

	return(sceneShader.isProgramLinked() && filterShader.isProgramLinked() && renderShader.isProgramLinked());
}

bool init()
{
	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

	// Create the VAO
	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);

	// Create buffers for our vertex data
	GLuint buffers[2];
	glGenBuffers(2, buffers);

	// Vertex coordinates buffer
	glBindBuffer(GL_ARRAY_BUFFER, buffers[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);
	glEnableVertexAttribArray(VSShaderLib::VERTEX_COORD_ATTRIB);
	glVertexAttribPointer(VSShaderLib::VERTEX_COORD_ATTRIB, 4, GL_FLOAT, 0, 0, 0);

	// Index buffer
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffers[1]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(faceIndex), faceIndex, GL_STATIC_DRAW);

	// Unbind the VAO
	glBindVertexArray(0);

	//surfRev.createSphere(0.75f, 20);
	//surfRev.createCone(1.f, 0.75f, 20);
	surfRev.createPawn();

	// Setup material properties
	VSResourceLib::Material materialArray[numObjects];

	srand(0);
	for(int i = 0; i < numObjects; i++)
	{
		for( int j = 0; j < 3; j++)
		{
			materialArray[i].diffuse[j] = rangeRandom(0.f, 1.f);
			materialArray[i].ambient[j] = materialArray[i].diffuse[j]/10.f;
			materialArray[i].specular[j] = 1.f;
		}
		materialArray[i].ambient[3] = 1.f;
		materialArray[i].diffuse[3] = 1.f;
		materialArray[i].specular[3] = 1.f;
		materialArray[i].shininess = rangeRandom(20.f, 150.f);
	}

	VSShaderLib::setBlock("Materials", materialArray);

	// Setup framebuffers
	static const GLenum renderBuffers[] = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1 };

	// Scene texture & framebuffer
	glGenFramebuffers(1, &renderFramebuffer);
	glBindFramebuffer(GL_FRAMEBUFFER, renderFramebuffer);
	glGenTextures(1, &sceneTexture);
	glBindTexture(GL_TEXTURE_2D, sceneTexture);
	glTexStorage2D(GL_TEXTURE_2D, 1, GL_RGBA16F, maxWindowWidth, maxWindowHeight);
	glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, sceneTexture, 0);

	// Highlight texture and scene depth texture
	glGenTextures(1, &highlightTexture);
	glBindTexture(GL_TEXTURE_2D, highlightTexture);
	glTexStorage2D(GL_TEXTURE_2D, 1, GL_RGBA16F, maxWindowWidth, maxWindowHeight);
	glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT1, highlightTexture, 0);
	glGenTextures(1, &depthTexture);
	glBindTexture(GL_TEXTURE_2D, depthTexture);
	glTexStorage2D(GL_TEXTURE_2D, 1, GL_DEPTH_COMPONENT32F, maxWindowWidth, maxWindowHeight);
	glFramebufferTexture(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, depthTexture, 0);
	glDrawBuffers(2, renderBuffers);
	
	// Filter textures
	glGenFramebuffers(2, filterFramebuffer);
	glGenTextures(2, filterTexture);
	for(int i = 0; i < 2; i++)
	{
		glBindFramebuffer(GL_FRAMEBUFFER, filterFramebuffer[i]);
		glBindTexture(GL_TEXTURE_2D, filterTexture[i]);
		glTexStorage2D(GL_TEXTURE_2D, 1, GL_RGBA16F, (i == 0) ? maxWindowHeight : maxWindowWidth, (i == 0) ? maxWindowWidth : maxWindowHeight);
		glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, filterTexture[i], 0);
		glDrawBuffers(1, renderBuffers);
	}
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	// Some GL settings
	glEnable(GL_CULL_FACE);
	glEnable(GL_MULTISAMPLE);

	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

	return true;
}

void initVSL()
{
//	Init VSML
	vsml = VSMathLib::getInstance();
	vsml->setUniformBlockName("Matrices");
	vsml->setUniformName(VSMathLib::VIEW, "m_view");
	vsml->setUniformName(VSMathLib::PROJECTION, "m_proj");
}


// ------------------------------------------------------------
//
// Main function
//

int main(int argc, char **argv)
{
//  GLUT initialization
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE|GLUT_RGBA|GLUT_MULTISAMPLE);

	glutInitContextVersion(3, 3);
	glutInitContextProfile(GLUT_CORE_PROFILE);
	glutInitContextFlags(GLUT_DEBUG);

	glutInitWindowPosition(100,100);
	glutInitWindowSize(512,512);
	glutCreateWindow("PV227 - Bloom");


//  Callback Registration
	glutDisplayFunc(renderScene);
	glutReshapeFunc(changeSize);
	glutIdleFunc(renderScene);

//	Mouse and Keyboard Callbacks
	glutKeyboardFunc(processKeys);
	glutMouseFunc(processMouseButtons);
	glutMotionFunc(processMouseMotion);

	glutMouseWheelFunc(mouseWheel);

//	Return from main loop
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

//	Init GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	printf("Vendor: %s\n", glGetString(GL_VENDOR));
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("Version: %s\n", glGetString(GL_VERSION));
	printf("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));
	int param;
	glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &param);
	if(param == GL_CONTEXT_CORE_PROFILE_BIT)
		printf("Context Profile: Core\n");
	else
		printf("Context Profile: Compatibility\n");

	if(!setupShaders())
		exit(1);

	if(!init())
		exit(1);

	initVSL();

//  GLUT main loop
	glutMainLoop();

	return(0);
}
